;;;
;;; hello.lisp
;;;
;;; The hello-world demo.
;;;

(cl:defpackage :hello-win32
  (:use :common-lisp :winapi :sb-alien)
  (:export #:do-it-to-it))

(in-package :hello-win32)

(defparameter *lambda* (make-array 2 :element-type '(unsigned-byte 16) :initial-contents '(#x3bb 0)))

(defun window-procedure (hwnd imsg wparam lparam)
  (cond ((= imsg wm_paint)
	 (with-alien ((ps (struct paintstruct))
		      (rect (struct rect)))
	   (let ((hdc (beginpaint hwnd (addr ps))))
	     (getclientrect hwnd (addr rect))
	     (drawtext hdc "Hello Lisp World!" -1 (addr rect)
		       (logior dt_singleline dt_center dt_vcenter))

	     ;; ---- 8< ----
	     ;; This nasty bit draws the lambda. Pretend it's not here.
	     (setf (slot rect 'bottom) (+ (slot rect 'top)
					  (ash (- (slot rect 'bottom)
						  (slot rect 'top)) -1)))
	     (alien-funcall
	      (extern-alien "DrawTextW" (function int hdc (* unsigned-short) int (* (struct rect)) unsigned-int))
	      hdc (sb-int::vector-sap *lambda*) -1 (addr rect) (logior dt_singleline dt_center dt_vcenter))
	     ;; ---- 8< ----

	     (endpaint hwnd (addr ps))))
	 0)
	#+(or)
	((= imsg wm_keyup)
	 (break)
	 0)
	((= imsg wm_destroy)
	 (postquitmessage 0)
	 0)
	(t
	 (defwindowproc hwnd imsg wparam lparam))))

(defparameter *window-procedure*
  (sb-alien::alien-callback
   (function unsigned-int hwnd unsigned-int unsigned-long unsigned-long)
   #'(lambda (x0 x1 x2 x3) (window-procedure x0 x1 x2 x3)) :stdcall))

;; Our hInstance is the load address of the runtime.
(defparameter *default-hinstance* #x400000)

(defun register-window-class ()
  (with-alien ((wndclass (struct wndclassex)))
    (setf (slot wndclass 'cbsize) 48)
    (setf (slot wndclass 'style) (logior cs_hredraw cs_vredraw))
    (setf (slot wndclass 'lpfnwndproc) (alien-sap *window-procedure*))
    (setf (slot wndclass 'cbclsextra) 0)
    (setf (slot wndclass 'cbwndextra) 0)
    (setf (slot wndclass 'hinstance) *default-hinstance*)
    (setf (slot wndclass 'hicon) (loadicon 0 idi_application))
    (setf (slot wndclass 'hcursor) (loadcursor 0 idc_arrow))
    (setf (slot wndclass 'hbrbackground) (getstockobject white_brush))
    (setf (slot wndclass 'lpszmenuname) nil)
    (setf (slot wndclass 'lpszclassname) "hello")
    (setf (slot wndclass 'hiconsm) (loadicon 0 idi_application))
    (registerclassex (addr wndclass))))

(defun unregister-window-class ()
  (unregisterclass "hello" *default-hinstance*))

(defun create-window ()
  (let ((hwnd (createwindow "hello" "SBCL/Win32 Hello Program"
			    ws_overlappedwindow
			    cw_usedefault cw_usedefault
			    cw_usedefault cw_usedefault
			    nil nil *default-hinstance* nil)))
    (showwindow hwnd sw_shownormal)
    (updatewindow hwnd)))

(defun message-loop ()
  (with-alien ((msg (struct msg)))
    (tagbody
     :head
       (when (zerop (getmessage (addr msg) nil 0 0))
	 (go :tail))
       (translatemessage (addr msg))
       (dispatchmessage (addr msg))
       (go :head)
     :tail
       (slot msg 'wparam))))

(defun do-it-to-it ()
  (register-window-class)
  (create-window)
  (message-loop)
  (unregister-window-class))

;;; EOF
